//////////////////////////////////////////////////////////////////////
// PSKGen.cpp: implementation of the CPSKGen class.
//
//////////////////////////////////////////////////////////////////////
//SoundCard/Wavefile signal generator program
// Copyright 1999.    Moe Wheatley AE4JY  <ae4jy@mindspring.com>
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//////////////////////////////////////////////////////////////////////
//

#include "stdafx.h"
//#include "WSCGen.h"
#include "PSKGen.h"
#include <math.h>
#include "tables.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

// local defines.................
#define DOPPLER_RANGE 500.0		//+\- range of doppler shift
#define DOPPLER_RATE 00.0		//Hz/Sec rate of doppler shift

#define PSK_CONTINUOUS 1		//comment out for one time sending

#define PHZ_0 0			//specify various signal phase states
#define PHZ_90 1
#define PHZ_180 2
#define PHZ_270 3
#define PHZ_OFF 4

#define SYM_NOCHANGE 0	//Stay the same phase
#define SYM_P90 1		//Plus 90  deg
#define SYM_P180 2		//Plus 180 deg
#define SYM_M90 3		//Minus 90 deg
#define SYM_OFF 4		//No output
#define SYM_ON 5		//constant output

#define QSIZE 1000				// input character Queue size

#define SYMBOL_RATE31 31.25		// 31.25 Symbols per Second
#define SYMBOL_RATE63 (SYMBOL_RATE31*2.)
#define SYMBOL_RATE125 (SYMBOL_RATE31*4.)
#define RAMP_SIZE 128			// number of envelope ramp steps per symbol

#define BPSK_RMS_CONSTANT 1.1247// Fudge factor to make average BPSK signal RMS = 1
#define QPSK_RMS_CONSTANT 1.1565// Fudge factor to make average QPSK signal RMS = 1
#define CW_RMS_CONSTANT 1.0		// Factor to make CW signal RMS = 1

#define TXOFF_CODE -1			// control codes that can be placed in the input
#define TXON_CODE -2			// queue for various functions
#define TXTOG_CODE -3

#define IMD31FREQ (SYMBOL_RATE31+(SYMBOL_RATE31/2.0))
#define IMD63FREQ (SYMBOL_RATE63+(SYMBOL_RATE63/2.0))
#define IMD125FREQ (SYMBOL_RATE125+(SYMBOL_RATE125/2.0))

// For the QPSK generator, rate 1/2 constraint length 5 
//   convolutional FEC coding is used.
// The generator polynomials used are:
//  g1(x) = x^4 + x^3           + 1		= 0x19
//  g0(x) = x^4	+       x^2 + x + 1		= 0x17
// 
//                                g1(x)
//              /----+--------+--------------------------+
//            /      |        |                          |
// symbol msb       ---      ---      ---      ---      ---
//                 | b4|<---| b3|<---| b2|<---| b1|<---| b0| <-- inverted data in
// symbol lsb       ---      ---      ---      ---      ---
//            \      |                 |        |        |
//              \-----+-----------------+--------+--------+
//                                g0(x)
//
// Lookup table to get symbol from non-inverted data stream
static const BYTE ConvolutionCodeTable[32] = 
{
	2, 1, 3, 0, 3, 0, 2, 1,
	0, 3, 1, 2, 1, 2, 0, 3,
	1, 2, 0, 3, 0, 3, 1, 2,
	3, 0, 2, 1, 2, 1, 3, 0 
};

//Ramp shape tables that are loaded with cosine shaped functions at init
static double PSKShapeTbl_Z[RAMP_SIZE];		// 0 
static double PSKShapeTbl_P[RAMP_SIZE];		// +1
static double PSKShapeTbl_M[RAMP_SIZE];		// -1
static double PSKShapeTbl_ZP[RAMP_SIZE];	// 0 to +1
static double PSKShapeTbl_PZ[RAMP_SIZE];	// +1 to 0
static double PSKShapeTbl_MZ[RAMP_SIZE];	// -1 to 0
static double PSKShapeTbl_PM[RAMP_SIZE];	// +1 to -1
static double PSKShapeTbl_MP[RAMP_SIZE];	// -1 to +1

struct PSKSTRUCT
{
	double* iptr;
	double* qptr;
	INT	next;
};
typedef PSKSTRUCT PSKSTATE;

//Lookup table for determining the next ramp shape depending on the
//  next symbol and the present output phase.
// indexing format is [symbol][presentPhase]
//returns the PSKSTRUCT containing the next phase and the I and Q
//   ramp table pointers.
static PSKSTATE PSKPhaseLookupTable[6][5]=
{
// SYMBOL = 0 = SYM_NOCHANGE
//   I ramp shape     Q ramp shape     Next Phase
	PSKShapeTbl_P, PSKShapeTbl_P, PHZ_0,	//present PHZ_0
	PSKShapeTbl_M, PSKShapeTbl_P, PHZ_90,	//present PHZ_90
	PSKShapeTbl_M, PSKShapeTbl_M, PHZ_180,	//present PHZ_180
	PSKShapeTbl_P, PSKShapeTbl_M, PHZ_270,	//present PHZ_270
	PSKShapeTbl_Z, PSKShapeTbl_Z, PHZ_OFF,	//present PHZ_OFF
// SYMBOL = 1 = SYM_P90 = Advance 90 degrees
//   I ramp shape     Q ramp shape     Next Phase
	PSKShapeTbl_PM, PSKShapeTbl_P, PHZ_90,	//present PHZ_0
	PSKShapeTbl_M, PSKShapeTbl_PM, PHZ_180,	//present PHZ_90
	PSKShapeTbl_MP, PSKShapeTbl_M, PHZ_270,	//present PHZ_180
	PSKShapeTbl_P, PSKShapeTbl_MP, PHZ_0,	//present PHZ_270
	PSKShapeTbl_ZP, PSKShapeTbl_ZP, PHZ_0,	//present PHZ_OFF
// SYMBOL = 2 = SYM_P180 = Advance 180 degrees
//   I ramp shape     Q ramp shape     Next Phase
	PSKShapeTbl_PM, PSKShapeTbl_PM, PHZ_180,//present PHZ_0
	PSKShapeTbl_MP, PSKShapeTbl_PM, PHZ_270,//present PHZ_90
	PSKShapeTbl_MP, PSKShapeTbl_MP, PHZ_0,	//present PHZ_180
	PSKShapeTbl_PM, PSKShapeTbl_MP, PHZ_90,	//present PHZ_270
	PSKShapeTbl_ZP, PSKShapeTbl_ZP, PHZ_0,	//present PHZ_OFF
// SYMBOL = 3 = SYM_M90	= retard 90 degrees
//   I ramp shape     Q ramp shape     Next Phase
	PSKShapeTbl_P, PSKShapeTbl_PM, PHZ_270,	//present PHZ_0
	PSKShapeTbl_MP, PSKShapeTbl_P, PHZ_0,	//present PHZ_90
	PSKShapeTbl_M, PSKShapeTbl_MP, PHZ_90,	//present PHZ_180
	PSKShapeTbl_PM, PSKShapeTbl_M, PHZ_180,	//present PHZ_270
	PSKShapeTbl_ZP, PSKShapeTbl_ZP, PHZ_0,	//present PHZ_OFF
// SYMBOL = 4 = SYM_OFF
//   I ramp shape     Q ramp shape     Next Phase
	PSKShapeTbl_PZ, PSKShapeTbl_PZ, PHZ_OFF,//present PHZ_0
	PSKShapeTbl_MZ, PSKShapeTbl_PZ, PHZ_OFF,//present PHZ_90
	PSKShapeTbl_MZ, PSKShapeTbl_MZ, PHZ_OFF,//present PHZ_180
	PSKShapeTbl_PZ, PSKShapeTbl_MZ, PHZ_OFF,//present PHZ_270
	PSKShapeTbl_Z, PSKShapeTbl_Z, PHZ_OFF,	//present PHZ_OFF
// SYMBOL = 4 = SYM_ON	
//   I ramp shape     Q ramp shape     Next Phase
	PSKShapeTbl_P, PSKShapeTbl_P, PHZ_0,	//present PHZ_0
	PSKShapeTbl_MP, PSKShapeTbl_P, PHZ_0,	//present PHZ_90
	PSKShapeTbl_MP, PSKShapeTbl_MP, PHZ_0,	//present PHZ_180
	PSKShapeTbl_P, PSKShapeTbl_MP, PHZ_0,	//present PHZ_270
	PSKShapeTbl_ZP, PSKShapeTbl_ZP, PHZ_0	//present PHZ_OFF
};
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPSKGen::CPSKGen()
{
	m_2PI = 8.0 * atan(1.0);		//make constant 2 PI
	m_pCharQueue = NULL;
	m_DopplerInc = 0.0;
	m_DopplerUP = TRUE;
}

CPSKGen::~CPSKGen()
{
	if(m_pCharQueue)
	{
		delete m_pCharQueue;
		m_pCharQueue = NULL;
	}

}

///////////++++++++++++++++++++++++++++++++++////////////////
///////////   P S K 3 1   G E N E R A T O R  ////////////////
///////////++++++++++++++++++++++++++++++++++////////////////

/////////////////////////////////////////////////////////////
//Initialize the PSK/CW generator with the sample frequncy Fs,
// the carrier frequency freq, and the PSK mode.
/////////////////////////////////////////////////////////////
void CPSKGen::InitPSK(INT Fs, double freq, INT mode)
{
	//mode bits 1:0 ---> 00==BPSK, 01==QPSK, 10==IQPSK, 11==CW
	//mode bits 3:2 ---> 00==31.25bps, 01==63bps, 10==125bps
	SetPSKFreq( Fs, freq,0,0,FALSE);	//set carrier frequency
	if( (mode&0x0C) == PSKSPD_31)
	{
		m_SymbolRate = SYMBOL_RATE31;
		m_IMDfreq = IMD31FREQ;
	}
	else if( (mode&0x0C) == PSKSPD_63)
	{
		m_SymbolRate = SYMBOL_RATE63;
		m_IMDfreq = IMD63FREQ;
	}
	else if( (mode&0x0C) == PSKSPD_125)
	{
		m_SymbolRate = SYMBOL_RATE125;
		m_IMDfreq = IMD125FREQ;
	}

	m_PSKmode = mode&0x03;			//save off local copy of mode
	switch(m_PSKmode)
	{
		case BPSK_MODE:
			m_RMSConstant = BPSK_RMS_CONSTANT;
			break;
		case QPSK_MODE:
		case IQPSK_MODE:
			m_RMSConstant = QPSK_RMS_CONSTANT;
			break;
		case CW_MODE:
			m_RMSConstant = CW_RMS_CONSTANT;
			break;
	}
	m_PSKSecPerSamp = 1.0/Fs;
	m_DeltaPerSecSec = m_2PI*m_PSKSecPerSamp*m_PSKSecPerSamp;
	m_PSKTime = 0;
//update envelope ramp every RAMP_SIZE times the symbol rate
	m_PSKPeriodUpdate = 1.0/(m_SymbolRate*RAMP_SIZE);	//symbol period
// Generate cosine ramp envelope lookup tables
	for( INT i=0; i<RAMP_SIZE; i++)
	{
		PSKShapeTbl_Z[i] = 0.0;
		PSKShapeTbl_P[i] = 1.0;
		PSKShapeTbl_M[i] = -1.0;
		PSKShapeTbl_PM[i] = cos( (double)i*m_2PI/(RAMP_SIZE*2) );
		PSKShapeTbl_MP[i] = -PSKShapeTbl_PM[i];

		if( i <64 )		//change to 64? from input from Vojtech OK1IAK
		{
			PSKShapeTbl_PZ[i] = cos( (double)i*m_2PI/(RAMP_SIZE*2) );
			PSKShapeTbl_MZ[i] = -PSKShapeTbl_PZ[i];
			PSKShapeTbl_ZP[i] = 0.0;
		}
		else
		{
			PSKShapeTbl_ZP[i] = -cos( (double)i*m_2PI/(RAMP_SIZE*2) );
			PSKShapeTbl_PZ[i] = 0.0;
			PSKShapeTbl_MZ[i] = 0.0;
		}

	}
	m_pPSKtxI = PSKShapeTbl_Z;
	m_pPSKtxQ = PSKShapeTbl_Z;
	m_pCharQueue = new INT[QSIZE];
	m_pHead = m_pTail = m_pCharQueue;
	m_PresentPhase = PHZ_OFF;
	m_TxShiftReg = 0;
	m_TxCodeWord = 0;
	CreateTestChars(m_PSKmode);	// put some stuff in character queue
	m_Finished = FALSE;
	m_DopplerRange = 0.0;
	m_DopplerRate = 0.0;

}

/////////////////////////////////////////////////////////////
// Call to change the psk31 carrier frequency
/////////////////////////////////////////////////////////////
void CPSKGen::SetPSKFreq(INT Fs, double freq, INT rate, INT range, BOOL idle)
{
	m_DopplerRange = (double)range;
	m_DopplerRate = (double)rate;
	if( (rate==0) || (range==0) )
	{
		m_DopplerInc = 0.0;
	}
	if(idle!=m_IdleOnly)
	{
		m_pHead = m_pTail = m_pCharQueue;
	}
	m_IdleOnly = idle;
	m_PSKPhaseInc = m_2PI * freq/(double)Fs;		//carrier frequency
	m_IMDPhaseInc1 = m_2PI * (freq + m_IMDfreq)/(double)Fs;
	m_IMDPhaseInc2 = m_2PI * (freq - m_IMDfreq)/(double)Fs;
}

/////////////////////////////////////////////////////////////
// generates n samples of psk31 waveform in pData with average RMS value of 1.0
/////////////////////////////////////////////////////////////
BOOL CPSKGen::CalcPSK(double *pData, INT n, double imdgn)
{
static double t = 0.0;
static double t1 = 0.0;
static double t2 = 0.0;
static BYTE Ramp = 0;
static double IMDgn = 0.0;
INT symbol;
INT i;
	for( i=0; i<n; i++ )
	{
		t = t + (m_PSKPhaseInc+m_DopplerInc);	// increment radian phase count
		t1 = t1 + (m_IMDPhaseInc1+m_DopplerInc);
		t2 = t2 + (m_IMDPhaseInc2+m_DopplerInc);
		pData[i] = m_RMSConstant*( m_pPSKtxI[Ramp]*sin( t ) + m_pPSKtxQ[Ramp]*cos( t ) );
 		pData[i] = pData[i] + IMDgn*sin( t1 ) + IMDgn*sin( t2 );
		m_PSKTime += m_PSKSecPerSamp;
		if( m_PSKTime >= m_PSKPeriodUpdate )//if time to update envelope ramp index
		{
			if( ++Ramp >= 128 )				// if time to update symbol
			{
				Ramp = 0;
				t = fmod(t,m_2PI);			//keep radian counter bounded
				t1 = fmod(t1,m_2PI);			//keep time bounded
				t2 = fmod(t2,m_2PI);			//keep time bounded
				switch( m_PSKmode )
				{
					case CW_MODE:
						symbol = GetNextCWSymbol();
						break;
					case BPSK_MODE:
						symbol = GetNextBPSKSymbol();
						break;
					case QPSK_MODE:
						symbol = GetNextQPSKSymbol();
						break;
					case IQPSK_MODE:
						symbol = GetNextQPSKSymbol();
						if(symbol==SYM_P90)		//rotate vectors the opposite way
							symbol = SYM_M90;
						else
							if(symbol==SYM_M90)
							symbol = SYM_P90;
						break;
				}
				if(symbol == SYM_P180)
					IMDgn = imdgn;
				else
					IMDgn = 0.0;
				//get new I/Q ramp tables and next phase
				m_pPSKtxI = PSKPhaseLookupTable[symbol][m_PresentPhase].iptr;
				m_pPSKtxQ = PSKPhaseLookupTable[symbol][m_PresentPhase].qptr;
				m_PresentPhase = PSKPhaseLookupTable[symbol][m_PresentPhase].next;
				if( m_pHead == m_pTail )	// if Character Queue is empty
				{
#ifdef PSK_CONTINUOUS
					CreateTestChars(m_PSKmode);	// refill queue if empty
#else
					m_Finished = TRUE;
#endif
				}
			}
			m_PSKTime -= m_PSKPeriodUpdate;	//keep time bounded
		}
		if(m_DopplerUP)
		{
			m_DopplerInc = m_DopplerInc + m_DopplerRate*m_DeltaPerSecSec;
			if( m_DopplerInc >= m_DopplerRange*m_2PI*m_PSKSecPerSamp )
				m_DopplerUP = FALSE;
		}
		else
		{
			m_DopplerInc = m_DopplerInc - m_DopplerRate*m_DeltaPerSecSec;
			if( m_DopplerInc <= -m_DopplerRange*m_2PI*m_PSKSecPerSamp )
				m_DopplerUP = TRUE;
		}
	}
	return m_Finished;
}

/////////////////////////////////////////////////////////////
//get next character from Queue. If got a new character in 'ch' then return TRUE
//else if empty return FALSE
/////////////////////////////////////////////////////////////
BOOL CPSKGen::GetChar(INT * ch) //get next character from Queue
{
BOOL result;
	if( m_pHead != m_pTail )	// if Queue has a character in it
	{
		m_pTail++;				// pre increment tail pointer
		if( m_pTail >= (m_pCharQueue+QSIZE) )
			m_pTail = m_pCharQueue;
		*ch = *m_pTail;
		result = TRUE;
	}
	else		// else is empty so return FALSE
		result = FALSE;
	return result;
}

/////////////////////////////////////////////////////////////
//put next symbol into symbol Queue ,if Queue full this thread will first
// wait for the consumer thread to remove a symbol.
/////////////////////////////////////////////////////////////
BOOL CPSKGen::PutChar( INT ch)
{
INT* pTempHead;
	pTempHead = m_pHead;
	pTempHead++;				// pre increment TempHead pointer
	if( pTempHead >= (m_pCharQueue+QSIZE) )
		pTempHead = m_pCharQueue;
	while( pTempHead == m_pTail )	// if Queue is full wait for it
	{
		Sleep(100);
	}
	m_pHead = pTempHead;		// update head ptr
	*m_pHead = ch;
	return TRUE;
}

/////////////////////////////////////////////////////////////
// called every symbol time to get next CW symbol and get the
// next character from the character Queue if no more symbols
// are left to send.
/////////////////////////////////////////////////////////////
BYTE CPSKGen::GetNextCWSymbol(void)
{
static BYTE lastsymb = SYM_OFF;
static INT CWState = 0;
static INT CWtimer;
BYTE symb;
INT ch;
	symb = lastsymb;		//use last symbol unless it needs to change
	if( (m_TxShiftReg == 0 ) && (CWState == 0) )
	{
		if( GetChar(&ch) )		//get next character to xmit
		{	
			if( ch >=0 )			//if is not a control code
			{
				ch &= 0xFF;		
				ch = (INT)toupper( (BYTE)ch );	//make upper case
				if( ch>=' ' && ch<='Z')
					m_TxShiftReg = CW_TABLE[ ch-' '];	//look up pattern
			}
			else					// is a control code
			{
				if( ch == TXON_CODE )
					symb = SYM_ON;
				else
					symb = SYM_OFF;
			}
		}
		CWState = 0;
	}
	switch( CWState )		// CW timing state machine
	{
		case 0:			//get next cw symbol state from pattern
			switch( m_TxShiftReg&0xC000 )
			{
				case 0x4000:	//dot
					CWState = 1;
					CWtimer = 2;
					symb = SYM_ON;
					break;
				case 0x8000:	//dash
					CWState = 1;
					CWtimer = 6;
					symb = SYM_ON;
					break;
				case 0xC000:	//inter char space 
					CWState = 2;
					CWtimer = 4;
					symb = SYM_OFF;
					break;
				default:
					symb = SYM_OFF;
					break;
			}
			m_TxShiftReg = m_TxShiftReg<<2;	//
			break;
		case 1:		//On time state
			if( --CWtimer <= 0 )
			{
				CWState = 2;
				CWtimer = 2;	//inter symbol time
				symb = SYM_OFF;
			}
			break;
		case 2:		//Off time state
			if( --CWtimer <= 0 )
				CWState = 0;
			break;
	}
	lastsymb = symb;
	return symb;
}

/////////////////////////////////////////////////////////////
// called every symbol time to get next BPSK symbol and get the
// next character from the character Queue if no more symbols
// are left to send.
/////////////////////////////////////////////////////////////
BYTE CPSKGen::GetNextBPSKSymbol(void)
{
static BOOL AddEndingZero = TRUE;
static BYTE lastsymb = SYM_OFF;
BYTE symb;
INT ch;
	symb = lastsymb;
	if( m_TxShiftReg == 0 )
	{
		if( AddEndingZero )			// if is end of code
		{
			symb = SYM_P180;		// end with a zero
			AddEndingZero = FALSE;
		}
		else
		{
			if( GetChar(&ch) )		//get next character to xmit
			{	
				if( ch >=0 )			//if is not a control code
				{						//get next VARICODE codeword to send
					m_TxShiftReg = VARICODE_TABLE[ ch&0xFF ];
					symb = SYM_P180;	//Start with a zero
				}
				else					// is a control code
					switch( ch )
					{
					case TXON_CODE:
						symb = SYM_ON;
						break;
					case TXTOG_CODE:
						symb = SYM_P180;
						break;
					case TXOFF_CODE:
						symb = SYM_OFF;
						break;
					}
			}
		}
	}
	else			// is not end of code word so send next bit
	{
		if( m_TxShiftReg&0x8000 )
			symb = SYM_NOCHANGE;
		else
			symb = SYM_P180;
		m_TxShiftReg = m_TxShiftReg<<1;	//point to next bit
		if( m_TxShiftReg == 0 )			// if at end of codeword
			AddEndingZero = TRUE;		// need to send a zero nextime
	}
	lastsymb = symb;
	return symb;
}

		
/////////////////////////////////////////////////////////////
// called every symbol time to get next QPSK symbol and get the
// next character from the character Queue if no more symbols
// are left to send.
/////////////////////////////////////////////////////////////
BYTE CPSKGen::GetNextQPSKSymbol(void)
{
static BOOL AddEndingZero = FALSE;
BYTE symb;
INT ch;
	symb = ConvolutionCodeTable[m_TxShiftReg&0x1F];	//get next convolution code
	m_TxShiftReg = m_TxShiftReg<<1;
	if( m_TxCodeWord == 0 )		//need to get next codeword
	{
		if( AddEndingZero )			//if need to add a zero
		{
			AddEndingZero = FALSE;	//end with a zero
		}
		else
		{
			if( GetChar(&ch) )		//get next character to xmit
			{	
				if( ch >=0 )			//if not a control code
				{						//get next VARICODE codeword to send
					m_TxCodeWord = VARICODE_TABLE[ ch&0xFF ];
				}
				else					//is a control code
				{
					switch( ch )
					{
					case TXON_CODE:
						symb = SYM_ON;
						break;
					case TXTOG_CODE:
						m_TxCodeWord = 0;
						break;
					case TXOFF_CODE:
						symb = SYM_OFF;
						break;
					}
				}
			}
			else
			{
				m_TxCodeWord = 0;
			}
		}
	}
	else
	{
		if(m_TxCodeWord&0x8000 )
		{
			m_TxShiftReg |= 1;
		}
		m_TxCodeWord = m_TxCodeWord<<1;
		if(m_TxCodeWord == 0)
			AddEndingZero = TRUE;	//need to add another zero
	}
	return symb;
}

/////////////////////////////////////////////////////////////
// place a bunch of characters in the character queue
/////////////////////////////////////////////////////////////
void CPSKGen::CreateTestChars(INT mode)
{
INT i, ch;
static INT IMDTimer = 0;
	if( mode == CW_MODE )
	{
		for( i=0; i<=100; i++ )
		{
			ch = rand()/668 + 41;	//get random character [* to Z]
			PutChar( (BYTE)ch );
			if( (i%5) == 0 )
				PutChar( ' ' );
		}
	}
	else
	{
		if(m_IdleOnly)
		{
			for( i=0; i<32; i++ )
				PutChar( TXTOG_CODE );
		}
		else
		{
			PutChar( 0x0D );			//send a CR LF
			PutChar( 0x0A );
			for( i=' '; i<='~'; i++ )	//send printable character set
			{
				PutChar( i );
//				PutChar( 'H' );
				if( ++IMDTimer > 10 )
				{
					IMDTimer = 0;
//					for(INT j=0; j<64; j++ )		//send 32 toggle codes
//						PutChar( TXTOG_CODE );
				}
			}
			for( i=0; i<32; i++ )		//send 32 toggle codes
				PutChar( TXTOG_CODE );
//			for( i=0; i<32; i++ )		//send steady carrier
//				PutChar( TXON_CODE );
//			for( i=0; i<32; i++ )		//send carrier off
//				PutChar( TXOFF_CODE );
		}
	}
}
///////////////////////////////////////////////////////////////////////
